#include "esp_camera.h"
#include "Arduino.h"
#include "SD.h"
#include "FS.h"

#include <WiFi.h>
#include <WebServer.h>
#include <ESPmDNS.h>

#include <ESP32_MailClient.h>

// Pin definition for CAMERA_MODEL_WROVER_KIT
#define PWDN_GPIO_NUM    -1
#define RESET_GPIO_NUM   -1
#define XCLK_GPIO_NUM    21
#define SIOD_GPIO_NUM    26
#define SIOC_GPIO_NUM    27

#define Y9_GPIO_NUM      35
#define Y8_GPIO_NUM      34
#define Y7_GPIO_NUM      39
#define Y6_GPIO_NUM      36
#define Y5_GPIO_NUM      19
#define Y4_GPIO_NUM      18
#define Y3_GPIO_NUM       5
#define Y2_GPIO_NUM       4
#define VSYNC_GPIO_NUM   25
#define HREF_GPIO_NUM    23
#define PCLK_GPIO_NUM    22

// Pin definition for SD Card
#define sd_sck  13
#define sd_mosi 15
#define sd_ss   12
#define sd_miso 14

WebServer server(80);
WebServer st_server(81);

const char *SSID = "aaaaaaaaaa";
const char *PASSWORD = "bbbbbbbbbb";

int cam_state[25]={0};
#define ope_stat          24

#define SMTP_HOST "smtp.gmail.com"
#define SMTP_PORT 465

/* The sign in credentials */
#define AUTHOR_EMAIL "cccccccc@gmail.com"
#define AUTHOR_PASSWORD "xxxxxxxxxxx"

/* Recipient's email*/
#define RECIPIENT_EMAIL "zzzzzzzzzz"

/* The SMTP Session object used for Email sending */
SMTPData smtpData;

void setup() 
{
    Serial.begin(115200);
    delay(100);

    Serial.println("Connecting to WiFi");
    WiFi.disconnect(true);
    WiFi.softAPdisconnect(true);
    delay(500);

    WiFi.mode(WIFI_STA);
    WiFi.begin(SSID, PASSWORD);
    delay(1000);

    // Try forever
    while (WiFi.status() != WL_CONNECTED) 
    {
        Serial.println("...Connecting to WiFi");
        delay(1000);
    }
    Serial.println("Connected");
    Serial.println(SSID);
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());

    if (MDNS.begin("esp32cam")) 
    {
        Serial.println("MDNS responder started");
    }

    init_cam();

    SPI.begin(sd_sck, sd_miso, sd_mosi, sd_ss);
    SD.begin(sd_ss);

    server.begin();
    server.on("/", handleRoot);
    server.onNotFound(handleWebRequests);
    Serial.println("HTTP server started");

    st_server.begin();
    st_server.on("/", st_handleRoot);

    read_State();
    cam_state[ope_stat] = 0;
}

void loop() 
{
    server.handleClient();
    st_server.handleClient();
}

void send_mail()
{
    //send email
    Serial.println("Sending email...");
    //Set the Email host, port, account and password
    smtpData.setLogin("smtp.gmail.com", 465, AUTHOR_EMAIL, AUTHOR_PASSWORD);
  
    //Set the sender name and Email
    smtpData.setSender("ESP32-CAM", AUTHOR_EMAIL);
  
    //Set Email priority or importance High, Normal, Low or 1 to 5 (1 is highest)
    smtpData.setPriority("Normal");

    //Set the subject
    smtpData.setSubject("Motion Detected - ESP32-CAM");
    
    //Set the message - normal text or html format
//    smtpData.setMessage("<div style=\"color:#003366;font-size:20px;\">Image captured and attached.</div>", true);
    smtpData.setMessage("Image captured and attached.", false);

    //Add recipients, can add more than one recipient
    smtpData.addRecipient(RECIPIENT_EMAIL);
  
    //Add attach files from SD card
    smtpData.addAttachFile("/data.jpg");
  
    //Set the storage types to read the attach files (SD is default)
    smtpData.setFileStorageType(MailClientStorageType::SD);
  
    smtpData.setSendCallback(sendCallback);
  
    //Start sending Email, can be set callback function to track the status
    if (!MailClient.sendMail(smtpData))
        Serial.println("Error sending Email, " + MailClient.smtpErrorReason());

    //Clear all data from Email object to free memory
    smtpData.empty();
}

/* Callback function to get the Email sending status */
void sendCallback(SendStatus msg)
{
    //Print the current status
    Serial.println(msg.info());

    //Do something when complete
    if (msg.success())
    {
        Serial.println("----------------");
    }
}

void init_cam()
{
    camera_config_t config;

    config.ledc_channel = LEDC_CHANNEL_0;
    config.ledc_timer = LEDC_TIMER_0;
    config.pin_d0 = Y2_GPIO_NUM;
    config.pin_d1 = Y3_GPIO_NUM;
    config.pin_d2 = Y4_GPIO_NUM;
    config.pin_d3 = Y5_GPIO_NUM;
    config.pin_d4 = Y6_GPIO_NUM;
    config.pin_d5 = Y7_GPIO_NUM;
    config.pin_d6 = Y8_GPIO_NUM;
    config.pin_d7 = Y9_GPIO_NUM;
    config.pin_xclk = XCLK_GPIO_NUM;
    config.pin_pclk = PCLK_GPIO_NUM;
    config.pin_vsync = VSYNC_GPIO_NUM;
    config.pin_href = HREF_GPIO_NUM;
    config.pin_sscb_sda = SIOD_GPIO_NUM;
    config.pin_sscb_scl = SIOC_GPIO_NUM;
    config.pin_pwdn = PWDN_GPIO_NUM;
    config.pin_reset = RESET_GPIO_NUM;
    config.xclk_freq_hz = 20000000;
    config.pixel_format = PIXFORMAT_JPEG;
    config.frame_size = FRAMESIZE_VGA; // FRAMESIZE_ + QVGA|CIF|VGA|SVGA|XGA|SXGA|UXGA
    config.jpeg_quality = 10;
    config.fb_count = 2;

    // Init Camera
    esp_camera_init(&config);
}

void read_State()
{
    sensor_t * s = esp_camera_sensor_get();

    cam_state[0]= s->status.framesize;
    cam_state[1]= s->status.quality;
    cam_state[2]= s->status.brightness;
    cam_state[3]= s->status.contrast;
    cam_state[4]= s->status.saturation;
    cam_state[5]= s->status.sharpness;
    cam_state[6]= s->status.special_effect;
    cam_state[7]= s->status.awb;
    cam_state[8]= s->status.awb_gain;
    cam_state[9]= s->status.wb_mode;
    cam_state[10]= s->status.aec;
    cam_state[11]= s->status.aec2;
    cam_state[12]= s->status.ae_level;
    cam_state[13]= s->status.agc;
    cam_state[14]= s->status.agc_gain;
    cam_state[15]= s->status.gainceiling;
    cam_state[16]= s->status.bpc;
    cam_state[17]= s->status.wpc;
    cam_state[18]= s->status.raw_gma;
    cam_state[19]= s->status.lenc;
    cam_state[20]= s->status.hmirror;
    cam_state[21]= s->status.vflip;
    cam_state[22]= s->status.dcw;
    cam_state[23]= s->status.colorbar;
}

void st_handleRoot() 
{
    WiFiClient client;
    camera_fb_t * fb = NULL;

    client = st_server.client();
    String response = "HTTP/1.1 200 OK\r\n";
    response += "Content-Type: multipart/x-mixed-replace; boundary=--frame\r\n\r\n";
    st_server.sendContent(response);
    while (1)
    {
      fb = esp_camera_fb_get();
      if (!client.connected())
      {
        esp_camera_fb_return(fb);
        break;
      }

      response = "--frame\r\n";
      response += "Content-Type: image/jpeg\r\n\r\n";
      st_server.sendContent(response);

      client.write(fb->buf, fb->len);
      st_server.sendContent("\r\n");
      esp_camera_fb_return(fb);

      if (!client.connected()) break;
    }
}

void handleRoot() 
{
    String buf,cmd;
    int a,b,fl,c_state[24];
    File dataFile;
    camera_fb_t * fb = NULL;
    sensor_t * s;

    fl=1;
    cmd=server.argName(0);
    switch(cmd.toInt())
    {
      case 1: // Get Still
              fb = esp_camera_fb_get();
              dataFile = SD.open("/data.jpg", FILE_WRITE);
              dataFile.write(fb->buf, fb->len); // payload (image), payload length
              dataFile.close();
              esp_camera_fb_return(fb);
              Serial.printf("Take a photo.\n");
              cam_state[ope_stat] = 2;
              break;

      case 2: //Set Camera Parameter
              buf=server.arg("2");
              a=buf.toInt();
              if(a < 0)
              {
                a = -a; a -= 10;
                cam_state[a] ^= 1; cam_state[a] &= 1;
              }
              else
              {
                a -= 10;
                b=0; while(buf[b] != ',') b ++;
                b ++;
                cmd=buf.substring(b);
                cam_state[a]=cmd.toInt();
              }
              s = esp_camera_sensor_get();
              switch(a)
              {
                case 0: s->set_framesize(s,(framesize_t)cam_state[0]); break;         //framesize
                case 1: s->set_quality(s,cam_state[1]); break;                        //quality
                case 2: s->set_brightness(s,cam_state[2]); break;                     //brightness
                case 3: s->set_contrast(s,cam_state[3]); break;                       //contrast
                case 4: s->set_saturation(s,cam_state[4]); break;                     //saturation
                case 5: s->set_sharpness(s,cam_state[5]); break;                      //sharpness
                case 6: s->set_special_effect(s,cam_state[6]); break;                 //special_effect
                case 7: s->set_whitebal(s,cam_state[7]); break;                       //awb
                case 8: s->set_awb_gain(s,cam_state[8]); break;                       //awb_gain
                case 9: s->set_wb_mode(s,cam_state[9]); break;                        //wb_mode
                case 10:  s->set_aec2(s,cam_state[10]); break;                        //aec
                case 11:  s->set_exposure_ctrl(s,cam_state[11]); break;               //aec2
                case 12:  s->set_ae_level(s,cam_state[12]); break;                    //ae_level
                case 13:  s->set_gain_ctrl(s,cam_state[13]); break;                   //agc
                case 14:  s->set_agc_gain(s,cam_state[14]); break;                    //agc_gain
                case 15:  s->set_gainceiling(s,(gainceiling_t)cam_state[15]); break;  //gainceiling
                case 16:  s->set_bpc(s,cam_state[16]); break;                         //bpc
                case 17:  s->set_wpc(s,cam_state[17]); break;                         //wpc
                case 18:  s->set_raw_gma(s,cam_state[18]); break;                     //raw_gma
                case 19:  s->set_lenc(s,cam_state[19]); break;                        //lenc
                case 20:  s->set_hmirror(s,cam_state[20]); break;                     //hmirror
                case 21:  s->set_vflip(s,cam_state[21]); break;                       //vflip
                case 22:  s->set_dcw(s,cam_state[22]); break;                         //dcw
                case 23:  s->set_colorbar(s,cam_state[23]); break;                    //colorba
              }
              break;

      case 3: // Send camera parameter
              cmd="";
              for(a=0; a<25; a++) cmd += (String(cam_state[a]) + ',');
              server.send(200, "text/plain", cmd);
              fl=0;
              break;

      case 4: //  Stream Start Stop
              if(cam_state[ope_stat] == 1) cam_state[ope_stat] = 0;
              else cam_state[ope_stat] = 1;
              break;
              
      case 5: //  Send mail
              send_mail();
              break;
    }

    if(fl)
      {
        dataFile = SD.open("/menu.html", FILE_READ);
        server.streamFile(dataFile,"text/html");
        dataFile.close();
      }
}

void handleWebRequests()
{
    String dataType = "text/plain";
    String path;
    File dataFile;
    camera_fb_t * fb;

    path = server.uri();
    if(path.endsWith(".txt")) dataType = "text/plain";
    else if(path.endsWith(".jpg")) dataType = "image/jpg";
    else if(path.endsWith(".css")) dataType = "text/css";
    else if(path.endsWith(".js")) dataType = "application/javascript";
    else if(path.endsWith(".png")) dataType = "image/png";
    else if(path.endsWith(".html")) dataType = "text/html";

    dataFile = SD.open(path.c_str(), "r");
    server.streamFile(dataFile, dataType);
    dataFile.close();
    delay(5);
}
